@@ -26,6 +26,7 @@ import java.util.ArrayList;
 import java.util.Arrays;
 import java.util.Collections;
 import java.util.Enumeration;
+import java.util.concurrent.ConcurrentHashMap;
 
 /**
  * A classloader which can delegate to multiple other classloaders without risk of deadlock.  A concurrent class loader
@@ -35,9 +36,7 @@ import java.util.Enumeration;
  *
  * @author <a href="mailto:david.lloyd@redhat.com">David M. Lloyd</a>
  */
-public abstract class ConcurrentClassLoader extends ClassLoader {
-
-    private static final ClassLoader definingLoader = ConcurrentClassLoader.class.getClassLoader();
+public abstract class ConcurrentClassLoader extends NamedClassLoader {
 
     private static final ThreadLocal<Boolean> GET_PACKAGE_SUPPRESSOR = new ThreadLocal<Boolean>();
 
@@ -65,22 +64,46 @@ public abstract class ConcurrentClassLoader extends ClassLoader {
      * @param parent the parent class loader
      */
     protected ConcurrentClassLoader(final ConcurrentClassLoader parent) {
-        super(parent == null ? ConcurrentClassLoader.class.getClassLoader() : parent);
-        if (getClassLoadingLock("$TEST$") == this) {
-            throw new Error("Cannot instantiate non-parallel subclass");
-        }
+        this(parent, null);
     }
 
     /**
      * Construct a new instance, using our class loader as the parent.
      */
     protected ConcurrentClassLoader() {
-        super(ConcurrentClassLoader.class.getClassLoader());
-        if (getClassLoadingLock("$TEST$") == this) {
+        this((String) null);
+    }
+
+    /**
+     * Construct a new instance with the given parent class loader, which must be a concurrent class loader, or {@code null}
+     * to create a root concurrent class loader.
+     *
+     * @param parent the parent class loader
+     * @param name the name of this class loader, or {@code null} if it is unnamed
+     */
+    protected ConcurrentClassLoader(final ConcurrentClassLoader parent, final String name) {
+        super(parent == null ? JDKSpecific.getPlatformClassLoader() : parent, name);
+        if (! JDKSpecific.isParallelCapable(this)) {
             throw new Error("Cannot instantiate non-parallel subclass");
         }
     }
 
+    /**
+     * Construct a new instance, using our class loader as the parent.
+     *
+     * @param name the name of this class loader, or {@code null} if it is unnamed
+     */
+    protected ConcurrentClassLoader(String name) {
+        super(JDKSpecific.getPlatformClassLoader(), name);
+        if (! JDKSpecific.isParallelCapable(this)) {
+            throw new Error("Cannot instantiate non-parallel subclass");
+        }
+    }
+
+    static Object getLockForClass(ConcurrentClassLoader cl, String name) {
+        return cl.getClassLoadingLock(name);
+    }
+
     /**
      * Loads the class with the specified binary name.  Equivalent to calling {@link #loadClass(String, boolean) loadClass(className, false)}.
      *
@@ -179,8 +202,7 @@ public abstract class ConcurrentClassLoader extends ClassLoader {
      */
     protected final Class<?> defineOrLoadClass(final String className, final byte[] bytes, int off, int len, ProtectionDomain protectionDomain) {
         try {
-            final Class<?> definedClass = defineClass(className, bytes, off, len, protectionDomain);
-            return definedClass;
+            return defineClass(className, bytes, off, len, protectionDomain);
         } catch (LinkageError e) {
             final Class<?> loadedClass = findLoadedClass(className);
             if (loadedClass != null) {
@@ -200,6 +222,21 @@ public abstract class ConcurrentClassLoader extends ClassLoader {
         return findClass(className, false, false);
     }
 
+    /**
+     * Implementation of {@link ClassLoader#findClass(String, String)}.
+     *
+     * @param moduleName the Java module name
+     * @param className the class name
+     * @return the result of {@code findClass(className, false, false)}
+     */
+    protected final Class<?> findClass(final String moduleName, final String className) {
+        try {
+            return findClass(className);
+        } catch (ClassNotFoundException e) {
+            return null;
+        }
+    }
+
     /**
      * Finds the resource with the given name.  The name of a resource is a {@code '/'}-separated path name that
      * identifies the resource.  If the resource name starts with {@code "java/"} then the parent class loader is used.
@@ -212,9 +249,7 @@ public abstract class ConcurrentClassLoader extends ClassLoader {
     public final URL getResource(final String name) {
         for (String s : Module.systemPaths) {
             if (name.startsWith(s)) {
-                // Whatever loads jboss-modules.jar should have it's classes accessible.
-                // This could even be the bootclasspath, in which case CL is null, and we prefer the system CL
-                return definingLoader != null ? definingLoader.getResource(name) : ClassLoader.getSystemResource(name);
+                return JDKSpecific.getSystemResource(name);
             }
         }
         return findResource(name, false);
@@ -232,7 +267,7 @@ public abstract class ConcurrentClassLoader extends ClassLoader {
     public final Enumeration<URL> getResources(final String name) throws IOException {
         for (String s : Module.systemPaths) {
             if (name.startsWith(s)) {
-                return definingLoader != null ? definingLoader.getResources(name) : ClassLoader.getSystemResources(name);
+                return JDKSpecific.getSystemResources(name);
             }
         }
         return findResources(name, false);
@@ -263,6 +298,19 @@ public abstract class ConcurrentClassLoader extends ClassLoader {
         return null;
     }
 
+    /**
+     * Find the resource with the given name in specified java module.
+     *
+     * @see #getResource(String)
+     *
+     * @param moduleName java module name
+     * @param name the resource name
+     * @return the resource URL
+     */
+    protected final URL findResource(final String moduleName, final String name) throws IOException {
+        return getResource(name);
+    }
+
     /**
      * Finds the resources with the given name and exported status.
      *
@@ -306,8 +354,7 @@ public abstract class ConcurrentClassLoader extends ClassLoader {
     }
 
     /**
-     * Returns an input stream for reading the specified resource.  If the resource starts with {@code "java/"}, then
-     * this method delegates to the parent class loader.  Otherwise, this method delegates to {@link #findResourceAsStream(String, boolean)}.
+     * Returns an input stream for reading the specified resource.  This method delegates to {@link #findResourceAsStream(String, boolean)}.
      *
      * @param name the resource name
      * @return the resource stream, or {@code null} if the resource is not found
@@ -315,7 +362,7 @@ public abstract class ConcurrentClassLoader extends ClassLoader {
     public final InputStream getResourceAsStream(final String name) {
         for (String s : Module.systemPaths) {
             if (name.startsWith(s)) {
-                return definingLoader != null ? definingLoader.getResourceAsStream(name) : ClassLoader.getSystemResourceAsStream(name);
+                return JDKSpecific.getSystemResourceAsStream(name);
             }
         }
         return findResourceAsStream(name, false);
@@ -345,7 +392,7 @@ public abstract class ConcurrentClassLoader extends ClassLoader {
         }
         for (String s : Module.systemPackages) {
             if (className.startsWith(s)) {
-                return definingLoader != null ? definingLoader.loadClass(className) : findSystemClass(className);
+                return JDKSpecific.getSystemClass(this, className);
             }
         }
         return performLoadClassUnchecked(className, exportsOnly, resolve);
@@ -363,7 +410,11 @@ public abstract class ConcurrentClassLoader extends ClassLoader {
         return findClass(className, exportsOnly, resolve);
     }
 
-    private final UnlockedReadHashMap<String, Package> packages = new UnlockedReadHashMap<String, Package>(64);
+    private final ConcurrentHashMap<String, Package> packages = new ConcurrentHashMap<>();
+
+    Class<?> findSystemClassInternal(String name) throws ClassNotFoundException {
+        return findSystemClass(name);
+    }
 
     /**
      * Load a package which is visible to this class loader.
@@ -441,7 +492,13 @@ public abstract class ConcurrentClassLoader extends ClassLoader {
             if (existing != null) {
                 return existing;
             }
-            Package pkg = super.definePackage(name, specTitle, specVersion, specVendor, implTitle, implVersion, implVendor, sealBase);
+            Package pkg;
+            try {
+                pkg = super.definePackage(name, specTitle, specVersion, specVendor, implTitle, implVersion, implVendor, sealBase);
+            } catch (final IllegalArgumentException iae) {
+                pkg = super.getPackage(name);
+                if (pkg == null) throw iae;
+            }
             existing = packages.putIfAbsent(name, pkg);
             return existing != null ? existing : pkg;
         } finally {
