@@ -265,7 +265,11 @@ public class ModuleLoader {
         return version == null ? module.getName() : module.getName() + "@" + version;
     }
 
-    static void installMBeanServer() {
+    /**
+     * Install JBoss Modules MBeans. WARNING: This method is not intended to be
+     * called by application code.
+     */
+    public static void installMBeanServer() {
         REG_REF.installReal();
     }
 
@@ -341,7 +345,7 @@ public class ModuleLoader {
                         }
                         final ModuleFinder finder = finders[idx++];
                         if (finder instanceof IterableModuleFinder) {
-                            nested = ((IterableModuleFinder) finder).iterateModules(baseName, recursive);
+                            nested = ((IterableModuleFinder) finder).iterateModules(baseName, recursive, ModuleLoader.this);
                         }
                     }
 
